# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import struct

from dstar import DSTARCallsign, DSTARModule
from stream import Packet, FixedPacket, DVHeaderPacket, DVFramePacket, DisconnectedError, ConnectionReceiveThread, Connection
from utils import or_valueerror, pad

class DPlusConnectPacket(FixedPacket):
    data = '\x05\x00\x18\x00\x01'

class DPlusLoginPacket(Packet):
    __slots__ = ['callsign', 'serial']

    def __init__(self, callsign, serial):
        self.callsign = callsign
        # dxrfd suggests uses "serial" to identify peer types.
        # Hotspots send "DV019999", DVAPs "AP", while everything
        # else is considered a "dongle".
        self.serial = serial

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 28)
        login, callsign, _, serial = struct.unpack('4s8s8s8s', data[4:])
        or_valueerror(login == '\x1c\xc0\x04\x00')
        return cls(callsign, serial)

    def to_data(self):
        return struct.pack('4s8s8s8s', '\x1c\xc0\x04\x00',
                                       str(self.callsign),
                                       '\x00\x00\x00\x00\x00\x00\x00\x00',
                                       pad(self.serial, 8))

class DPlusLoginOKPacket(FixedPacket):
    data = '\x08\xc0\x04\x00OKRW'

class DPlusLoginBusyPacket(FixedPacket):
    data = '\x08\xc0\x04\x00BUSY'

class DPlusLoginFailPacket(FixedPacket):
    data = '\x08\xc0\x04\x00FAIL'

class DPlusDisconnectPacket(FixedPacket):
    data = '\x05\x00\x18\x00\x00'

class DPlusKeepAlivePacket(FixedPacket):
    data = '\x03\x60\x00'

class DPlusHeaderPacket(Packet):
    __slots__ = ['dv_header']

    def __init__(self, dv_header):
        self.dv_header = dv_header

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 58)
        or_valueerror(data[:2] == '\x3a\x80')
        dv_header = DVHeaderPacket.from_data(data[2:])
        return cls(dv_header)

    def to_data(self):
        return '\x3a\x80' + self.dv_header.to_data()

class DPlusFramePacket(Packet):
    __slots__ = ['dv_frame']

    def __init__(self, dv_frame):
        self.dv_frame = dv_frame

    @classmethod
    def from_data(cls, data):
        if len(data) == 29:
            or_valueerror(data[:2] == '\x1d\x80')
            dv_frame = DVFramePacket.from_data(data[2:])
            return cls(dv_frame)
        elif len(data) == 32:
            or_valueerror(data[:2] == '\x20\x80')
            dv_frame = DVFramePacket.from_data(data[2:29])
            if not dv_frame.is_last:
                dv_frame.packet_id |= 64
            return cls(dv_frame)
        else:
            raise ValueError

    def to_data(self):
        if not self.dv_frame.is_last:
            return '\x1d\x80' + self.dv_frame.to_data()

        data = '\x20\x80' + \
               self.dv_frame.to_data()[:15] + \
               '\x55\xc8\x7a\x55\x55\x55\x55\x55\x55\x55\x55\x55\x25\x1a\xc6' # XXX Why?
        return data[:8] + '\x81' + data[9:]

class DPlusConnectionRecieveThread(ConnectionReceiveThread):
    def _process(self, data):
        try:
            packet = DPlusFramePacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received dvframe packet from stream %s%s', packet.dv_frame.stream_id, ' (last)' if packet.dv_frame.is_last else '')
            return packet

        try:
            packet = DPlusHeaderPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received dvheader packet from stream %s', packet.dv_header.stream_id)
            return packet

        try:
            packet = DPlusConnectPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received connect packet')
            return packet

        try:
            packet = DPlusLoginOKPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received login ok packet')
            return packet

        try:
            packet = DPlusLoginBusyPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received login busy packet')
            return packet

        try:
            packet = DPlusLoginFailPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received login fail packet')
            return packet

        try:
            packet = DPlusDisconnectPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received disconnect packet')
            return packet # XXX Request or reply?

        try:
            packet = DPlusKeepAlivePacket.from_data(data)
        except ValueError:
            pass
        else:
            # self.logger.debug('received keepalive packet')
            keepalive_packet = DPlusKeepAlivePacket()
            self.sock.write(keepalive_packet.to_data())
            return

        self.logger.warning('unknown data received')

class DPlusConnection(Connection):
    DEFAULT_PORT = 20001

    def __init__(self, callsign, module, reflector_callsign, reflector_module, reflector_address):
        Connection.__init__(self, callsign, module, reflector_callsign, reflector_module, reflector_address)
        self.receive_thread = DPlusConnectionRecieveThread(self.sock)

    def _connect(self, timeout=3):
        self.write(DPlusConnectPacket())
        packet = self._read(timeout, [DPlusConnectPacket])
        if not packet:
            return False

        self.write(DPlusLoginPacket(self.callsign, ''))
        packet = self._read(timeout, [DPlusLoginOKPacket, DPlusLoginBusyPacket, DPlusLoginFailPacket])
        if packet and isinstance(packet, DPlusLoginOKPacket):
            return True
        return False

    def _disconnect(self, timeout=3):
        self.write(DPlusDisconnectPacket())
        return True if self._read(timeout, [DPlusDisconnectPacket]) else False

    def read(self, timeout=3):
        packet = self._read(timeout, [DPlusHeaderPacket, DPlusFramePacket]) 
        if not packet:
            return None
        if isinstance(packet, DPlusHeaderPacket):
            return packet.dv_header
        if isinstance(packet, DPlusFramePacket):
            return packet.dv_frame

    def write(self, packet):
        if isinstance(packet, DVHeaderPacket):
            packet = DPlusHeaderPacket(packet)
        elif isinstance(packet, DVFramePacket):
            packet = DPlusFramePacket(packet)
        return self.sock.write(packet.to_data())



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import struct

from stream import DVHeaderPacket, DVFramePacket
from utils import or_valueerror

class DVToolFile(object):
    def __init__(self, name):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.debug('initialized with name %s', name)

        self.name = name
        self.f = None

    def open(self):
        or_valueerror(self.f is None)
        try:
            self.f = open(self.name, 'a+b')
        except Exception as e:
            self.logger.error('can not open file: %s', str(e))
            return False
        self.logger.debug('opened file %s', self.name)
        return True

    def close(self):
        if self.f:
            self.f.close()
        self.f = None
        self.logger.debug('closed file %s', self.name)

    def __enter__(self):
        if not self.open():
            raise Exception('can not open file %s' % self.name)
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def write(self, stream):
        self.f.seek(0)
        self.f.truncate()

        self.f.write('DVTOOL' + struct.pack('<I', len(stream)))
        for i, packet in enumerate(stream):
            self.f.write(struct.pack('<H', 56 if i == 0 else 27) + packet.to_data())
        self.logger.info('wrote a stream of %s packets in %s', len(stream), self.name)

    def read(self):
        self.f.seek(0)

        stream = []
        magic, count = struct.unpack('<6sI', self.f.read(10))
        or_valueerror(magic == 'DVTOOL')
        for i in xrange(count):
            size, = struct.unpack('<H', self.f.read(2))
            data = self.f.read(size)
            if i == 0:
                or_valueerror(size == 56)
                packet = DVHeaderPacket.from_data(data)
            else:
                or_valueerror(size == 27)
                packet = DVFramePacket.from_data(data)
            stream.append(packet)
        self.logger.info('read a stream of %s packets from %s', len(stream), self.name)
        return stream



# Copyright (C) 2017 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import struct

from dstar import DSTARCallsign, DSTARModule
from stream import Packet, FixedPacket, DVHeaderPacket, DVFramePacket, DisconnectedError, ConnectionReceiveThread, Connection
from utils import or_valueerror

class DExtraConnectPacket(Packet):
    __slots__ = ['src_callsign', 'src_module', 'dest_module', 'revision']

    def __init__(self, src_callsign, src_module, dest_module, revision):
        self.src_callsign = src_callsign
        self.src_module = src_module
        self.dest_module = dest_module
        self.revision = revision

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 11)
        src_callsign, src_module, dest_module, src_revision = struct.unpack('8sccB', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) != ' ')
        if src_revision == 11:
            revision = 1
        elif (str(src_callsign).startswith('XRF')):
            revision = 2
        else:
            revision = 0
        return cls(src_callsign, src_module, dest_module, revision)

    def to_data(self):
        return struct.pack('8sccB', str(self.src_callsign),
                                    str(self.src_module),
                                    str(self.dest_module),
                                    11 if self.revision == 1 else 0) # XXX What about revisions 0 and 2?

class DExtraConnectAckPacket(Packet):
    __slots__ = ['src_callsign', 'src_module', 'dest_module', 'revision']

    def __init__(self, src_callsign, src_module, dest_module, revision):
        self.src_callsign = src_callsign
        self.src_module = src_module
        self.dest_module = dest_module
        self.revision = revision

    @classmethod
    def from_data(cls, data):
        # Same size with disconnect
        # if (len(data) == 11):
        #     raise NotImplementedError

        or_valueerror(len(data) == 14)
        src_callsign, src_module, dest_module, ack = struct.unpack('8scc4s', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) != ' ')
        or_valueerror(ack == 'ACK\x00')
        return cls(src_callsign, src_module, dest_module, 1) # XXX Could it be revision 0?

    @classmethod
    def from_connect_packet(cls, connect_packet):
        return cls(connect_packet.src_callsign,
                   connect_packet.src_module,
                   connect_packet.dest_module,
                   connect_packet.revision)

    def to_data(self):
        if self.revision == 2:
            return struct.pack('8sccB', str(self.src_callsign),
                                        str(self.dest_module),
                                        str(self.src_module),
                                        0)

        return struct.pack('8scc4s', str(self.src_callsign),
                                     str(self.src_module),
                                     str(self.dest_module),
                                     'ACK\x00')

class DExtraConnectNackPacket(Packet):
    __slots__ = ['src_callsign', 'src_module', 'dest_module']

    def __init__(self, src_callsign, src_module, dest_module):
        self.src_callsign = src_callsign
        self.src_callsign = src_module
        self.dest_module = dest_module

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 14)
        src_callsign, src_module, dest_module, nack = struct.unpack('8scc4s', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) != ' ')
        or_valueerror(nack == 'NAK\x00')
        return cls(src_callsign, src_module, dest_module)

    @classmethod
    def from_connect_packet(cls, connect_packet):
        return cls(connect_packet.src_callsign,
                   connect_packet.src_module,
                   connect_packet.dest_module)

    def to_data(self):
        return struct.pack('8scc4s', str(self.src_callsign),
                                     str(self.src_module),
                                     str(self.dest_module),
                                     'NAK\x00')

class DExtraDisconnectPacket(Packet):
    __slots__ = ['src_callsign', 'src_module']

    def __init__(self, src_callsign, src_module):
        self.src_callsign = src_callsign
        self.src_module = src_module

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 11)
        src_callsign, src_module, dest_module, _ = struct.unpack('8sccB', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) == ' ')
        return cls(src_callsign, src_module)

    def to_data(self):
        return struct.pack('8sc2s', str(self.src_callsign),
                                    str(self.src_module),
                                    '  ')

class DExtraDisconnectAckPacket(FixedPacket):
    data = 'DISCONNECTED'

class DExtraKeepAlivePacket(Packet):
    __slots__ = ['src_callsign']

    def __init__(self, src_callsign):
        self.src_callsign = src_callsign

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 9)
        src_callsign, _ = struct.unpack('8sc', data)
        src_callsign = DSTARCallsign(src_callsign)
        return cls(src_callsign) # XXX Get module?

    def to_data(self):
        return str(self.src_callsign) + '\x00' # XXX Send module?

class DExtraConnectionRecieveThread(ConnectionReceiveThread):
    def __init__(self, sock, callsign):
        ConnectionReceiveThread.__init__(self, sock)
        self.callsign = callsign

    def _process(self, data):
        try:
            packet = DVFramePacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received dvframe packet from stream %s%s', packet.stream_id, ' (last)' if packet.is_last else '')
            return packet

        try:
            packet = DVHeaderPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received dvheader packet from stream %s', packet.stream_id)
            return packet

        try:
            packet = DExtraConnectAckPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received connect ack packet')
            return packet

        try:
            packet = DExtraConnectNackPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received connect nack packet')
            return packet

        try:
            packet = DExtraDisconnectPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received disconnect packet from %s', packet.src_callsign)
            raise DisconnectedError

        try:
            packet = DExtraDisconnectAckPacket.from_data(data)
        except ValueError:
            pass
        else:
            self.logger.debug('received disconnect ack packet')
            return packet

        try:
            packet = DExtraKeepAlivePacket.from_data(data)
        except ValueError:
            pass
        else:
            # self.logger.debug('received keepalive packet from %s', packet.src_callsign)
            keepalive_packet = DExtraKeepAlivePacket(self.callsign)
            self.sock.write(keepalive_packet.to_data())
            return

        self.logger.warning('unknown data received')

class DExtraConnection(Connection):
    DEFAULT_PORT = 30001

    def __init__(self, callsign, module, reflector_callsign, reflector_module, reflector_address):
        Connection.__init__(self, callsign, module, reflector_callsign, reflector_module, reflector_address)
        self.receive_thread = DExtraConnectionRecieveThread(self.sock, self.callsign)

    def _connect(self, timeout=3):
        self.write(DExtraConnectPacket(self.callsign, self.module, self.reflector_module, 1))
        packet = self._read(timeout, [DExtraConnectAckPacket, DExtraConnectNackPacket])
        if packet and isinstance(packet, DExtraConnectAckPacket):
            return True
        return False

    def _disconnect(self, timeout=3):
        self.write(DExtraDisconnectPacket(self.callsign, self.module))
        return True if self._read(timeout, [DExtraDisconnectAckPacket]) else False



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import struct
import time
import Queue

from dstar import DSTARHeader, DSTARFrame
from network import UDPClientSocket
from utils import or_valueerror, StoppableThread

class Packet(object): # Abstract
    __slots__ = ['data']

    def __init__(self, data):
        self.data = data

    @classmethod
    def from_data(cls, data):
        return cls(data)

    def to_data(self):
        return data

class FixedPacket(Packet):
    __slots__ = []

    def __init__(self):
        pass

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == len(cls.data))
        or_valueerror(data == cls.data)
        return cls()

    def to_data(self):
        return self.__class__.data

class DVHeaderPacket(Packet):
    __slots__ = ['band_1', 'band_2', 'band_3', 'stream_id', 'dstar_header']

    def __init__(self, band_1, band_2, band_3, stream_id, dstar_header):
        self.band_1 = band_1
        self.band_2 = band_2
        self.band_3 = band_3
        self.stream_id = stream_id
        self.dstar_header = dstar_header

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 56)
        or_valueerror(data[:4] == 'DSVT')
        or_valueerror(data[4] == '\x10')
        or_valueerror(data[8] == '\x20')
        band_1, band_2, band_3, stream_id = struct.unpack('<BBBH', data[9:14])
        dstar_header = DSTARHeader.from_data(data[15:])
        return cls(band_1, band_2, band_3, stream_id, dstar_header)

    def to_data(self):
        return ('DSVT\x10\x00\x00\x00\x20' +
                struct.pack('<BBBH', self.band_1, self.band_2, self.band_3, self.stream_id) +
                '\x80' +
                self.dstar_header.to_data())

class DVFramePacket(Packet):
    __slots__ = ['band_1', 'band_2', 'band_3', 'stream_id', 'packet_id', 'dstar_frame']

    def __init__(self, band_1, band_2, band_3, stream_id, packet_id, dstar_frame):
        self.band_1 = band_1
        self.band_2 = band_2
        self.band_3 = band_3
        self.stream_id = stream_id
        self.packet_id = packet_id
        self.dstar_frame = dstar_frame

    @property
    def is_last(self):
        return self.packet_id & 64 != 0

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 27)
        or_valueerror(data[:4] == 'DSVT')
        or_valueerror(data[4] == '\x20')
        or_valueerror(data[8] == '\x20')
        band_1, band_2, band_3, stream_id, packet_id = struct.unpack('<BBBHB', data[9:15])
        dstar_frame = DSTARFrame.from_data(data[15:])
        return cls(band_1, band_2, band_3, stream_id, packet_id, dstar_frame)

    def to_data(self):
        return ('DSVT\x20\x00\x00\x00\x20' +
                struct.pack('<BBBHB', self.band_1, self.band_2, self.band_3, self.stream_id, self.packet_id) +
                self.dstar_frame.to_data())

class DisconnectedError(Exception):
    pass

class ConnectionReceiveThread(StoppableThread):
    def __init__(self, sock):
        self.logger = logging.getLogger(self.__class__.__name__)

        StoppableThread.__init__(self, name=self.__class__.__name__)
        self._sleep_period = 0.01

        self.sock = sock
        self.queue = Queue.Queue()

    def _process(self, data): # Abstract
        if not data:
            raise DisconnectedError
        return Packet.from_data(data)

    def loop(self):
        while True: # While there is data to read from the socket
            data = self.sock.read()
            if not data:
                return

            try:
                packet = self._process(data)
                if packet:
                    self.queue.put(packet)
            except DisconnectedError:
                self.queue.put(None)

class Connection(object):
    def __init__(self, callsign, module, reflector_callsign, reflector_module, reflector_address):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.debug('initialized with callsign %s module %s reflector callsign %s reflector_module %s reflector_address %s', callsign, module, reflector_callsign, reflector_module, reflector_address)

        self.callsign = callsign
        self.module = module
        self.reflector_callsign = reflector_callsign
        self.reflector_module = reflector_module
        self.reflector_address = reflector_address

        self.sock = UDPClientSocket(self.reflector_address)
        self.receive_thread = ConnectionReceiveThread(self.sock)
        self.disconnected = False

    def _read(self, timeout=3, expected_packet_classes=None):
        step = 0.01
        clock = time.time()
        limit = clock + timeout
        while (clock < limit):
            try:
                packet = self.receive_thread.queue.get(True, step)
                if not packet:
                    self.disconnected = True
                    raise DisconnectedError
                if expected_packet_classes == None:
                    return packet
                for cls in expected_packet_classes:
                    if isinstance(packet, cls):
                        return packet
            except Queue.Empty:
                pass
            clock = time.time()
        return None

    def _connect(self): # Abstract
        return True

    def _disconnect(self): # Abstract
        return True

    def open(self):
        try:
            self.sock.open()
        except Exception as e:
            self.logger.error('can not open UDP socket: %s', str(e))
            return False
        self.logger.info('connected to reflector %s at address %s', self.reflector_callsign, self.reflector_address)

        self.receive_thread.start()
        self.disconnected = False

        return self._connect()

    def close(self):
        while not self.receive_thread.queue.empty():
            self.receive_thread.queue.get()

        if not self.disconnected:
            self._disconnect()

        self.receive_thread.join()

        self.sock.close()
        self.logger.info('disconnected from reflector %s at address %s', self.reflector_callsign, self.reflector_address)

    def __enter__(self):
        if not self.open():
            raise Exception('can not open connection to %s' % (self.reflector_address,))
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def read(self, timeout=3):
        return self._read(timeout, [DVHeaderPacket, DVFramePacket])

    def write(self, packet):
        return self.sock.write(packet.to_data())



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import argparse
import logging
import wave

import pydv.codec2

from dstar import DSTARHeader, DSTARFrame, DSTARCallsign, DSTARSuffix
from stream import DVHeaderPacket, DVFramePacket
from dvtool import DVToolFile

def dv_encoder():
    parser = argparse.ArgumentParser(description='D-STAR encoder. Encodes samples into streams.')
    parser.add_argument('-v', '--verbose', default=False, action='store_true', help='enable debug output')
    parser.add_argument('-m', '--mode', default='3200', help='Codec 2 mode (2400 or 3200)')
    parser.add_argument('-f', '--fec', default=False, action='store_true', help='enable FEC')
    parser.add_argument('input', help='name of file to encode (WAV format)')
    parser.add_argument('output', help='name of file to write (DVTool format)')
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG if args.verbose else logging.INFO)
    logger = logging.getLogger(os.path.basename(sys.argv[0]))

    if args.mode not in ('2400', '3200'):
        logger.error('mode can be either 2400 or 3200')
        sys.exit(1)
    if args.fec:
        logger.error('FEC is not implemented')
        sys.exit(1)

    wavef = wave.open(args.input, 'r')
    if (wavef.getnchannels() != 1 or wavef.getsampwidth() != 2 or wavef.getframerate() != 8000):
        logger.error('input file must have 1 channel, 16 bits/sample, and 8000 samples/sec')
        sys.exit(1)

    version = 0x01
    if args.mode == '2400':
        version |= 0x02
    # if args.fec:
    #     version &= 0x04
    dstar_header = DSTARHeader(0,
                               0,
                               version,
                               DSTARCallsign('NOCALL'),
                               DSTARCallsign('NOCALL'),
                               DSTARCallsign('NOCALL'),
                               DSTARCallsign('NOCALL'),
                               DSTARSuffix('    '))
    header = DVHeaderPacket(0, 0, 0, 0, dstar_header)
    stream = [header]
    packet_id = 0

    mode = pydv.codec2.MODE_2400 if (args.mode == '2400') else pydv.codec2.MODE_3200
    state = pydv.codec2.init_state(mode)
    while True:
        data = wavef.readframes(160)
        if len(data) < 160:
            break

        dvcodec = pydv.codec2.encode(state, data)
        dstar_frame = DSTARFrame(dvcodec, '\x55\x2d\x16' if (packet_id % 21 == 0) else '')
        packet = DVFramePacket(0, 0, 0, 0, packet_id % 21, dstar_frame)
        stream.append(packet)

        packet_id += 1

    # Mark last packet
    packet = stream[-1]
    packet.packet_id |= 64

    wavef.close()

    with DVToolFile(args.output) as f:
        f.write(stream)

def main():
    dv_encoder()

if __name__ == '__main__':
    dv_encoder()



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import argparse
import logging

from dstar import DSTARCallsign, DSTARModule
from dextra import DExtraConnection
from dplus import DPlusConnection
from stream import DisconnectedError, DVHeaderPacket, DVFramePacket
from network import NetworkAddress
from dvtool import DVToolFile

def dv_recorder():
    parser = argparse.ArgumentParser(description='D-STAR recorder. Connects to reflector and records traffic.')
    parser.add_argument('-v', '--verbose', default=False, action='store_true', help='enable debug output')
    parser.add_argument('-p', '--protocol', default='auto', help='network protocol (dextra, dplus, or auto)')
    parser.add_argument('callsign', help='your callsign')
    parser.add_argument('reflector', help='reflector\'s callsign')
    parser.add_argument('module', help='reflector\'s module')
    parser.add_argument('address', help='reflector\'s hostname or IP address')
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG if args.verbose else logging.INFO)
    logger = logging.getLogger(os.path.basename(sys.argv[0]))

    try:
        callsign = DSTARCallsign(args.callsign)
        reflector_callsign = DSTARCallsign(args.reflector)
        reflector_module = DSTARModule(args.module)
        if args.protocol == 'dextra':
            connection_class = DExtraConnection
        elif args.protocol == 'dplus':
            connection_class = DPlusConnection
        elif args.protocol == 'auto':
            connection_class = DPlusConnection if str(reflector_callsign).startswith('REF') else DExtraConnection
        else:
            raise ValueError
        reflector_address = NetworkAddress(args.address, connection_class.DEFAULT_PORT)
    except ValueError:
        parser.print_help()
        sys.exit(1)

    try:
        stream_id = None
        with connection_class(callsign, DSTARModule(' '), reflector_callsign, reflector_module, reflector_address) as conn:
            try:
                while True:
                    packet = conn.read()
                    if isinstance(packet, DVHeaderPacket):
                        if packet.stream_id != stream_id:
                            stream_id = packet.stream_id
                            stream = [packet]
                    elif isinstance(packet, DVFramePacket):
                        if packet.stream_id == stream_id:
                            stream.append(packet)
                            if packet.is_last:
                                with DVToolFile('%s.dvtool' % stream_id) as f:
                                    f.write(stream)
                                stream_id = None
                    else:
                        pass
            except (DisconnectedError, KeyboardInterrupt):
                pass
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)

def main():
    dv_recorder()

if __name__ == '__main__':
    dv_recorder()



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

def or_valueerror(condition):
    if not condition:
        raise ValueError

def pad(s, count, padding='\x00'):
    return s + ((count - len(s)) * padding)

def resolve(host):
    import socket

    try:
        socket.inet_aton(host)
        return host
    except socket.error:
        pass

    try:
        return socket.gethostbyname(host)
    except:
        raise ValueError

# Based on: https://www.oreilly.com/library/view/python-cookbook/0596001673/ch06s03.html

import threading

class StoppableThread(threading.Thread):
    def __init__(self, name='StoppableThread'):
        self._stop_event = threading.Event()
        self._sleep_period = 1.0

        threading.Thread.__init__(self, name=name)

    def loop(self):
        raise NotImplementedError

    def run(self):
        while not self._stop_event.isSet():
            self.loop()
            self._stop_event.wait(self._sleep_period)

    def join(self, timeout=None):
        self._stop_event.set()
        threading.Thread.join(self, timeout)



# Copyright (C) 2017 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import socket
import select
import logging

from collections import namedtuple

from utils import or_valueerror, resolve

_NetworkAddress = namedtuple('NetworkAddress', ['host', 'port'])
class NetworkAddress(_NetworkAddress):
    def __str__(self):
        return '%s:%s' % (self.host, self.port)

class UDPClientSocket(object):
    def __init__(self, remote_address, local_address=None):
        if local_address is None:
            local_address = NetworkAddress('0.0.0.0', 0)

        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.debug('initialized with remote %s local %s', remote_address, local_address)

        try:
            self.remote_address = NetworkAddress(resolve(remote_address.host), remote_address.port)
        except ValueError:
            self.logger.error('cannot find address for host %s', remote_address.host)
            raise
        self.local_address = local_address

        self.sock = None

    def open(self):
        or_valueerror(self.sock is None)

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(self.local_address)
        self.logger.debug('socket opened')

    def close(self):
        if self.sock is not None:
            self.sock.close()
            self.sock = None
        self.logger.debug('socket closed')

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def read(self, length=1024):
        or_valueerror(self.sock)

        # Check that the recvfrom() won't block
        readable, writable, exceptional = select.select([self.sock], [], [], 0) # Return immediately
        if not readable:
            return None

        data, address = self.sock.recvfrom(length)
        self.logger.debug('read %d bytes from %s: %s', len(data), address, repr(data))

        # Check if the data is for us, only check the IP address now (used to include port number too)
        if self.remote_address.host != address[0]:
            return None;

        return data

    def write(self, data):
        or_valueerror(self.sock)

        self.logger.debug('write %d bytes to %s: %s', len(data), self.remote_address, repr(data))
        length = self.sock.sendto(data, self.remote_address)
        self.logger.debug('wrote %d bytes', length)

        if length != len(data):
            return False
        return True

if __name__ == '__main__':
    import threading

    from time import sleep

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG)

    def server():
        logger = logging.getLogger('server')
    
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        server_address = NetworkAddress('0.0.0.0', 20000)
        sock.bind(server_address)
        logger.debug('server listening at %s', server_address)

        while True:
            data, address = sock.recvfrom(256)
            logger.debug('echoing %d bytes back to %s: %s', len(data), address, data)
            length = sock.sendto(data, address)
            logger.debug('echoed %d bytes', length)

    def client():
        sleep(1)
        data = 'test'
        with UDPClientSocket(NetworkAddress('127.0.0.1', 20000)) as udp_socket:
            udp_socket.write(data)
            while True:
                if udp_socket.read(256) == data:
                    break
                sleep(0.1)

    t = threading.Thread(target=client)
    t.start()

    try:
        server()
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python

import setuptools

def long_description():
    with open('README.md', 'r') as f:
        return f.read()

setuptools.setup(
    name='pydv',
    version='0.4',
    author='Antony Chazapis',
    author_email='chazapis@gmail.com',
    description='D-STAR library and utilities in Python',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/chazapis/pydv',
    license='GPLv2',
    packages=['pydv'],
    entry_points={'console_scripts': ['dv-recorder=pydv.recorder:main',
                                      'dv-player=pydv.player:main',
                                      'dv-encoder=pydv.encoder:main',
                                      'dv-decoder=pydv.decoder:main',
                                      'dv-transcoder=pydv.transcoder:main']},
    ext_modules=[setuptools.Extension(name='pydv.mbelib',
                                      sources=['pydv/mbelib.c'],
                                      libraries=['mbe']),
                 setuptools.Extension(name='pydv.codec2',
                                      sources=['pydv/codec2.c'],
                                      libraries=['codec2'])],
    classifiers=['Environment :: Console',
                 'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: Communications :: Ham Radio']
)



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import argparse
import logging
import random

from dstar import DSTARCallsign, DSTARSuffix, DSTARModule
from dextra import DExtraConnection
from dplus import DPlusConnection
from stream import DisconnectedError, DVHeaderPacket, DVFramePacket
from network import NetworkAddress
from dvtool import DVToolFile

def dv_player():
    parser = argparse.ArgumentParser(description='D-STAR player. Connects to reflector and plays back recordings.')
    parser.add_argument('-v', '--verbose', default=False, action='store_true', help='enable debug output')
    parser.add_argument('-p', '--protocol', default='auto', help='network protocol (dextra, dplus, or auto)')
    parser.add_argument('callsign', help='your callsign')
    parser.add_argument('reflector', help='reflector\'s callsign')
    parser.add_argument('module', help='reflector\'s module')
    parser.add_argument('address', help='reflector\'s hostname or IP address')
    parser.add_argument('input', help='name of file to play back (DVTool format)')
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG if args.verbose else logging.INFO)
    logger = logging.getLogger(os.path.basename(sys.argv[0]))

    try:
        callsign = DSTARCallsign(args.callsign)
        reflector_callsign = DSTARCallsign(args.reflector)
        reflector_module = DSTARModule(args.module)
        if args.protocol == 'dextra':
            connection_class = DExtraConnection
        elif args.protocol == 'dplus':
            connection_class = DPlusConnection
        elif args.protocol == 'auto':
            connection_class = DPlusConnection if str(reflector_callsign).startswith('REF') else DExtraConnection
        else:
            raise ValueError
        reflector_address = NetworkAddress(args.address, connection_class.DEFAULT_PORT)
    except ValueError:
        parser.print_help()
        sys.exit(1)

    try:
        with DVToolFile(args.input) as f:
            stream = f.read()
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)

    header = stream[0]
    header.dstar_header.my_callsign = callsign
    header.dstar_header.my_suffix = DSTARSuffix('    ')
    header.dstar_header.ur_callsign = DSTARCallsign('CQCQCQ')
    header.dstar_header.repeater_1_callsign = DSTARCallsign(str(reflector_callsign)[:7] + str(reflector_module))
    header.dstar_header.repeater_2_callsign = DSTARCallsign(str(reflector_callsign)[:7] + 'G')
    stream_id = random.getrandbits(16)

    try:
        with connection_class(callsign, DSTARModule(' '), reflector_callsign, reflector_module, reflector_address) as conn:
            try:
                for packet in stream:
                    packet.stream_id = stream_id
                    conn.write(packet)
            except (DisconnectedError, KeyboardInterrupt):
                pass
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)

def main():
    dv_player()

if __name__ == '__main__':
    dv_player()
